#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <termios.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <sys/time.h>

#include "config.h"
#include "COM1939.h"
#include "CANInterface.h"

const int RS_RXDATA = 4;
const int RS_REPSTATUS = 9;

// J1939 Constants
// ------------------------------------------------------------------------
const long PGN_RequestMessage = 0x00EA00;           // 59904dec
const long PGN_GlobalRequestMessage = 0x00EAFF;

// COM1939 internal functions ---------------------------------------------
void SendFilterMessage(long lPGN, int fileCOMPort, bool bAdd);
byte ComputeCheckSum(byte pMsg[], int nStart, int nEnd);

// CAN Controller Interface -----------------------------------------------
static CANInterface CAN;

COM1939::COM1939()
{

}

COM1939::~COM1939()
{

}

// ------------------------------------------------------------------------
// J1939 Initialization
// ------------------------------------------------------------------------
int COM1939::Initialize(char* sCOMPort)
{
    // Initialize the CAN connection
    strcpy(sPort, sCOMPort);
    int nResult = CAN.Initialize(&fileCOMPort, sCOMPort, &bCommunicationError, &nErrNum, &bErrorOpenCOMPort, &bErrorCOMPortSettings);

    // Reset the J1939 converter
    CAN.SendRESETMessage(fileCOMPort, CANBAUDRATE_250);

    // Return the result
    return nResult;

}// end Initialize

// ------------------------------------------------------------------------
// J1939 Protocol Operation
// ------------------------------------------------------------------------
// Returncode: Message return codes according to CANInterface.h
//
int COM1939::Receive(long* lPGN, int* nPriority, int* nSourceAddress, int* nDestAddress, byte nData[], int* nDataLen,
                     bool *bTimeStamp, byte nTimeStamp[])
{
    // Declarations
    Msg pMessage;

    // Default setting
    *lPGN = 0;
    *nPriority = 0;
    *nSourceAddress = NULLADDRESS;
    *nDestAddress = NULLADDRESS;
    *nDataLen = 0;
    *bTimeStamp = false;

    // Get the jCOM1939 heartbeat
    bjCOM1939Heartbeat = CAN.bHeartbeat;

    // Check for a received message
    int nMsgType = CAN.Receive(&pMessage, fileCOMPort);

    if(nMsgType == RX_Message || nMsgType == RX_TSMessage)
    {
        // Get the PGN
        *lPGN = ((long)pMessage.Data[RXTX_IDX_PGNMSB]) << 16;
        *lPGN |= ((long)pMessage.Data[RXTX_IDX_PGN2ND]) << 8;
        *lPGN |= ((long)pMessage.Data[RXTX_IDX_PGNLSB]);

        // Get more parameters
        *nPriority = (int)pMessage.Data[RXTX_IDX_PRIORITY];
        *nSourceAddress = (int)pMessage.Data[RXTX_IDX_SRCADDR];
        *nDestAddress = (int)pMessage.Data[RXTX_IDX_DESTADDR];

         // Correct the data length according to message type
        if(nMsgType == RX_TSMessage)
            *nDataLen = pMessage.DataLen - RXTX_TOTALOVERHEAD - 8;  // Cut off the protocol overhead plus timestamp
        else
            *nDataLen = pMessage.DataLen - RXTX_TOTALOVERHEAD;      // Cut off the protocol overhead

        // Copy the data
        int nPointer = 0;
        int nIndex = 0;
        for(nIndex = RXTX_IDX_DATASTART; nIndex < RXTX_IDX_DATASTART + (int)*nDataLen; nIndex++)
            nData[nPointer++] = pMessage.Data[nIndex];

        // Copy the timestamp
        nPointer = nIndex;
        if(nMsgType == RX_TSMessage)
        {
            for(nIndex = 0; nIndex < 8; nIndex++)
                nTimeStamp[nIndex] = pMessage.Data[nPointer++];

            *bTimeStamp = true;

        }// end if

    }// end if

    // Return the message type
    return nMsgType;

}// end Receive

// ------------------------------------------------------------------------
// J1939 Transmit
// ------------------------------------------------------------------------
int COM1939::Transmit(int nPriority, long lPGN, int nSourceAddress, int nDestAddress, byte nData[], int nDataLen)
{
    // Declarations
    Msg pMessage;
    int nMsgLen = nDataLen + TOTAL_OVERHEAD;   // Actual message length plus overhead and checksum

    // Fill the TX_1939 message header
    pMessage.Data[RXTX_IDX_MSGSTART] = MSG_TOKEN_START;
    pMessage.Data[RXTX_IDX_MSGID] = MSG_ID_TX;
    pMessage.Data[RXTX_IDX_PGNMSB] = 0;                                     // PGN MSB
    pMessage.Data[RXTX_IDX_PGN2ND] = (byte)((lPGN & (long)0x00FF00) >> 8);  // PGN 2nd byte
    pMessage.Data[RXTX_IDX_PGNLSB] = (byte)(lPGN & (long)0x0000FF);         // PGN LSB
    pMessage.Data[RXTX_IDX_DESTADDR] = (byte)nDestAddress;
    pMessage.Data[RXTX_IDX_SRCADDR] = (byte)nSourceAddress;
    pMessage.Data[RXTX_IDX_MSGLENMSB] = (byte)(nMsgLen >> 8);
    pMessage.Data[RXTX_IDX_MSGLENLSB] = (byte)nMsgLen;
    pMessage.Data[RXTX_IDX_PRIORITY] = (byte)nPriority;

    // Determine the checksum
    // We need the message without stuff bytes to calculate the checksum
    Msg pMsg;
    int nCounter = 0;

    // Copy the protocol overhead
    for(int nIndex = 0; nIndex < RXTX_IDX_DATASTART; nIndex++)
    pMsg.Data[nCounter++] = pMessage.Data[nIndex];

    // Copy the actual data without stuff bytes
    for(int nIndex = 0; nIndex < nDataLen; nIndex++)
    pMsg.Data[nCounter++] = nData[nIndex];

    // Finally, create the checksum
    byte nChecksum = ComputeCheckSum(pMsg.Data, 1, nCounter - 1);
    pMsg.Data[nCounter] = nChecksum;

    // Fill the TX_1939 message data including byte stuffing
    int nPointer = RXTX_IDX_MSGLENMSB;    // Points to beginning of stuffing range

    for (int nIndex = 1; nIndex < nCounter + 1; nIndex++)
    {
          switch (pMsg.Data[nIndex])
          {
              case MSG_TOKEN_START:
                  pMessage.Data[nPointer++] = MSG_TOKEN_ESC;
                  pMessage.Data[nPointer++] = MSG_START_STUFF;
                  break;

              case MSG_TOKEN_ESC:
                  pMessage.Data[nPointer++] = MSG_TOKEN_ESC;
                  pMessage.Data[nPointer++] = MSG_ESC_STUFF;
                  break;

              default:
                  pMessage.Data[nPointer++] = pMsg.Data[nIndex];
                  break;

          }// end switch

    }// end for

    // Process and fill the checksum
    pMessage.DataLen = nPointer;

    // Transmit the message
    return CAN.Transmit(pMessage, fileCOMPort);

}// end Transmit

// ------------------------------------------------------------------------
// J1939 Termination
// ------------------------------------------------------------------------
void COM1939::Terminate()
{
    // Reset the gateway
    CAN.SendRESETMessage(fileCOMPort, CANBAUDRATE_250);

    // Close the serial port
    CAN.ClosePort(fileCOMPort);

}// end Terminate

// ------------------------------------------------------------------------
// J1939 Set Protocol Parameters per SET message
// ------------------------------------------------------------------------
void COM1939::SetProtocolParameters(byte* nNAME, byte nSA, byte nSABottom, byte nSATop, byte nOpMode, bool bReqRSMessage)
{
    CAN.SendSETMessage(nNAME, nSA, nSABottom, nSATop, nOpMode, bReqRSMessage, fileCOMPort);

}// end SetProtocolParameters

//-FUNCTION-----------------------------------------------------------------
// Routine     : ComputeCheckSum
// Description : Computes the checksum of a message and inserts it
// -------------------------------------------------------------------------
byte ComputeCheckSum(byte pMsg[], int nStart, int nEnd)
{
    // Declarations
    int nCheckSum = 0;
    byte nResult = 0;

    // Create the checksum
    for (int nIndex = nStart; nIndex <= nEnd; nIndex++)
        nCheckSum += pMsg[nIndex];
    nCheckSum = ~nCheckSum; // Negation does not work with type byte
    nResult = (byte)nCheckSum;
    nResult += 1;

    return nResult;

}// end ComputeCheckSum

// Gateway management functions
// ------------------------------------------------------------------------
void COM1939::GetVersion(char* sSWVersion, char* sHWVersion)
{
    strcpy(sSWVersion, CAN.sSWVersion);
    strcpy(sHWVersion, CAN.sHWVersion);
}

void COM1939::ResetGateway(int nCANBaudRate)
{
    // Reset the J1939 converter
    CAN.SendRESETMessage(fileCOMPort, nCANBaudRate);
}

void COM1939::SendMessageMode(byte nMsgMode)
{
    CAN.SendMessageMode(fileCOMPort, nMsgMode);
}

void COM1939::AddFilter(long lPGN)
{
    CAN.SendFilterMessage(lPGN, fileCOMPort, true);
}

void COM1939::DelFilter(long lPGN)
{
    CAN.SendFilterMessage(lPGN, fileCOMPort, false);
}

void COM1939::AddFilterX(long lPGN, int nSampleRate)
{
    CAN.SendFilterMessageX(lPGN, nSampleRate, fileCOMPort);
}

void COM1939::RequestStatus(void)
{
    CAN.SendRequestMessage(fileCOMPort, MSG_ID_RS);
}

int COM1939::GetStatus(int* nSA)
{
    *nSA = CAN.nReportedSourceAddress;
    return CAN.nReportedAddrClaimStatus;
}

// COM1939.cpp


